#include <engine/e_server_interface.h>
#include <game/generated/g_protocol.hpp>

#include <game/server/gamecontext.hpp>
#include <game/server/gamemodes/exp/exp.hpp>
#include "door.hpp"



LASER_DOOR::LASER_DOOR(vec2 pos, int type, DOOR * r)
: ENTITY(NETOBJTYPE_LASER)
{
	this->pos = pos;
	ref = r;
	
	dir = vec2(0, 0);
	if(type == DOOR_TYPE_VERTICAL)
		dir = vec2(0, 1);
	else if(type == DOOR_TYPE_HORIZONTAL)
		dir = vec2(1, 0);
	
	create();
	
	game.world.insert_entity(this);
}


void LASER_DOOR::create()
{
	vec2 to = pos + dir*10000.0f;
	vec2 org_to = to;
	
	if(col_intersect_line(pos, to, NULL, &to, false))
	{
		//intersected
		
		from = pos;
		pos = to;
		
		vec2 temp_pos = pos;
		vec2 temp_dir = dir*4.0f;
		
		move_point(&temp_pos, &temp_dir, 1.0f, 0);
		pos = temp_pos;
		dir = normalize(temp_dir);
	}
	else
	{
		from = pos;
		pos = to;
	}
}

void LASER_DOOR::destroy()
{
	ref->create_laser = true;
	ref->laser = NULL;
	ENTITY::destroy();
}

void LASER_DOOR::reset()
{
	game.world.destroy_entity(this);
}

void LASER_DOOR::tick()
{
}

void LASER_DOOR::snap(int snapping_client)
{
	if(networkclipped(snapping_client))
		return;
	
	if(game.players[snapping_client]->get_character() && game.players[snapping_client]->get_character()->door_open())
		return;
	
	NETOBJ_LASER *obj = (NETOBJ_LASER *)snap_new_item(NETOBJTYPE_LASER, id, sizeof(NETOBJ_LASER));
	obj->x = (int)pos.x;
	obj->y = (int)pos.y;
	obj->from_x = (int)from.x;
	obj->from_y = (int)from.y;
	obj->start_tick = server_tick()-1;
}
