#include <engine/e_server_interface.h>
#include <game/generated/g_protocol.hpp>

#include <game/server/gamecontext.hpp>
#include <game/server/gamemodes/exp/exp.hpp>
#include "laser.hpp"



LASER::LASER(vec2 pos, vec2 direction, float start_energy, int owner, bool turret, bool freezer)
: ENTITY(NETOBJTYPE_LASER)
{
	is_turret = turret;
	turret_collision = false;
	is_freezer = freezer;
	
	this->pos = pos;
	this->owner = owner;
	energy = start_energy;
	dir = direction;
	bounces = 0;
	do_bounce();
	
	game.world.insert_entity(this);
}


bool LASER::hit_character(vec2 from, vec2 to)
{
	vec2 at;
	CHARACTER * hit = NULL;
	if(is_turret)
		hit = game.world.intersect_character(pos, to, 0.0f, at, NULL);
	else
		hit = game.world.intersect_character(pos, to, 0.0f, at, game.get_player_char(owner));
	if(!hit)
		return false;

	this->from = from;
	pos = at;
	energy = -1;
	
	if(is_turret)
	{
		if(!hit->player->is_bot())
			hit->take_damage(vec2(0, 0), tuning.turret_laser_damage, hit->player->client_id, WEAPON_WORLD);
	}
	else if(is_freezer)
	{
		if(!game.controller->is_friendly_fire(owner, hit->player->client_id))
			hit->freeze();
	}
	else
		hit->take_damage(vec2(0, 0), tuning.laser_damage, owner, WEAPON_RIFLE);
	
	return true;
}

void LASER::do_bounce()
{
	eval_tick = server_tick();
	
	if(energy < 0 || turret_collision)
	{
		game.world.destroy_entity(this);
		return;
	}
	
	vec2 to = pos + dir*energy;
	vec2 org_to = to;
	
	bool check_doors = false;
	if(owner >= 0 && game.players[owner])
	{
		if(game.players[owner]->get_character())
			check_doors = !game.players[owner]->get_character()->door_open();
		else
			check_doors = true;
	}
	
	if(col_intersect_line(pos, to, NULL, &to, check_doors))
	{
		if(!hit_character(pos, to))
		{
			// intersected
			from = pos;
			pos = to;

			vec2 temp_pos = pos;
			vec2 temp_dir = dir*4.0f;
			
			move_point(&temp_pos, &temp_dir, 1.0f, 0);
			pos = temp_pos;
			dir = normalize(temp_dir);
			
			energy -= distance(from, pos) + tuning.laser_bounce_cost;
			bounces++;
			
			if(bounces > tuning.laser_bounce_num || is_turret)
				energy = -1;
				
			game.create_sound(pos, SOUND_RIFLE_BOUNCE);
		}
	}
	else
	{
		if(!hit_character(pos, to))
		{
			from = pos;
			pos = to;
			energy = -1;
		}
	}
}
	
void LASER::reset()
{
	game.world.destroy_entity(this);
}

void LASER::tick()
{
	//TURRETS COLLISION
	if(game.controller->isEXP() && !is_turret && game.players[owner] && !game.players[owner]->is_bot()
		&& server_tick() > turret_hit_timer)
	{
		for(int b = 0; b < MAX_TURRETS; b++)
		{
			TURRET * t = &( ((GAMECONTROLLER_EXP *)game.controller)->turrets[b] );
			if(!t->used || t->dead)
				continue;
			
			float closest_len = distance(from, pos) * 100.0f;
			vec2 intersect_pos = closest_point_on_line(from, pos, t->pos);
			float len = distance(t->pos, intersect_pos);
			if(len < 32.0f && len < closest_len)
			{
				if(is_freezer)
					((GAMECONTROLLER_EXP *)game.controller)->freeze_turret(b);
				else
					((GAMECONTROLLER_EXP *)game.controller)->hit_turret(b, owner, 5);
				pos = t->pos;
				turret_collision = true;
			}
		}
	}
	
	if(turret_collision)
		turret_hit_timer = (float)server_tick() + 0.1f*server_tickspeed();
	if(server_tick() > eval_tick+(server_tickspeed()*tuning.laser_bounce_delay)/1000.0f)
		do_bounce();

}

void LASER::snap(int snapping_client)
{
	if(networkclipped(snapping_client))
		return;

	NETOBJ_LASER *obj = (NETOBJ_LASER *)snap_new_item(NETOBJTYPE_LASER, id, sizeof(NETOBJ_LASER));
	obj->x = (int)pos.x;
	obj->y = (int)pos.y;
	obj->from_x = (int)from.x;
	obj->from_y = (int)from.y;
	obj->start_tick = eval_tick;
}
