#include <engine/e_server_interface.h> //server_tick, server_tickspeed
#include <engine/e_config.h> //sv_max_clients

#include <game/server/gamecontext.hpp>
#include <game/server/player.hpp>
#include <game/server/entities/pickup.hpp>
#include "exp.hpp"
#include "bots.hpp"



void GAMECONTROLLER_EXP::tick_bots()
{
	//REMOVE DEAD BOTS
	for(int b = config.sv_max_clients; b < MAX_CLIENTS; b++)
	{
		if(game.players[b] && game.players[b]->must_remove_bot)
			remove_bot(b, true);
	}
	
	
	//CHECK FOR NOBODY
	for(int b = config.sv_max_clients; b < MAX_CLIENTS; b++)
	{
		PLAYER * p = game.players[b];
		if(!p || !p->get_character())
			continue;
		if(p->bot_level == 4)
			continue;
		
		bool nobody = true;
		for(int i = 0; i < config.sv_max_clients; i++)
		{
			if(game.players[i] && !p->get_character()->networkclipped(i))
			{
				nobody = false;
				break;
			}
		}
		
		if(nobody)
		{
			if(p->nobody_timer == 0)
			{
				p->nobody_timer = (float)server_tick() + 10.0f*server_tickspeed();
			}
			else
			{
				if((float)server_tick() > p->nobody_timer)
					remove_bot(b, false);
			}
		}
		else
			p->nobody_timer = 0;
	}
	
	
	//CHECK FOR BOT SPAWNS
	for(int l = 0; l < BOT_LEVELS; l++)
	{
		for(int s = 0; s < num_bot_spawns[l]; s++)
		{
			if(bot_can_spawn() == -1)
				break;
			
			BOT_SPAWN * spawn = &bot_spawns[l][s];
			if(spawn->spawned)
				continue;
			if((float)server_tick() < spawn->respawn_timer + tuning.respawn_timer*server_tickspeed())
				continue;
			
			for(int p = 0; p < config.sv_max_clients; p++)
			{
				if(!game.players[p] || !game.players[p]->get_character())
					continue;
				
				if(distance(game.players[p]->get_character()->pos, spawn->pos) < 700.0f)
				{
					bot_spawn(spawn);
					break;
				}
			}
		}
	}
	
	
	//TICK BOSS
	if(boss.exist)
	{
		if(boss.spawn.spawned)
		{
			if(distance(game.players[boss.client_id]->get_character()->pos, boss.spawn.pos) > tuning.boss_distancelimit)
			{
				game.players[boss.client_id]->get_character()->teleport(boss.spawn.pos);
			}
			
			if(boss.shield_active)
			{
				if((float)server_tick() > boss.regen_timer)
				{
					if(game.players[boss.client_id]->get_character()->health < game.players[boss.client_id]->max_health())
					{
						game.players[boss.client_id]->get_character()->health++;
						boss.regen_timer = (float)server_tick() + 1.0f*server_tickspeed();
					}
				}
				
				vec2 boss_pos = game.players[boss.client_id]->get_character()->pos;
				vec2 boss_vel = game.players[boss.client_id]->get_character()->core.vel;
				
				if(boss.shield_icons[0])
					boss.shield_icons[0]->pos = (boss_pos-boss_vel*1)*0.1f + boss.shield_icons[0]->pos*0.9f;
				if(boss.shield_icons[1])
					boss.shield_icons[1]->pos = (boss_pos-boss_vel*4)*0.1f + boss.shield_icons[1]->pos*0.9f;
				if(boss.shield_icons[2])
					boss.shield_icons[2]->pos = (boss_pos-boss_vel*7)*0.1f + boss.shield_icons[2]->pos*0.9f;
			}
			else
			{
				if((float)server_tick() > boss.shield_timer)
				{
					boss.shield_active = true;
					boss.shield_health = 18;
					for(int s = 0; s < 3; s++)
					{
						boss.shield_icons[s] = new PICKUP(1, 0);
						boss.shield_icons[s]->make_boss_shield();
						boss.shield_icons[s]->pos = game.players[boss.client_id]->get_character()->pos;
					}
				}
			}
		}
		else
		{
			for(int p = 0; p < config.sv_max_clients; p++)
			{
				if(!game.players[p] || !game.players[p]->get_character())
					continue;
				if(game.players[p]->game_exp.boss_killer) //he has already killed the boss
					continue;
				
				if(distance(game.players[p]->get_character()->pos, boss.spawn.pos) < 700.0f)
				{
					bot_spawn(&boss.spawn);
					break;
				}
			}
		}
	}
}





int GAMECONTROLLER_EXP::bot_can_spawn()
{
	for(int p = config.sv_max_clients; p < MAX_CLIENTS; p++)
		if(!game.players[p]) return p;
	return -1;
}


void GAMECONTROLLER_EXP::bot_spawn(BOT_SPAWN * spawn)
{
	ENTITY * ents[1] = {0};
	int num_ents = game.world.find_entities(spawn->pos, 28.0f, ents, 1, NETOBJTYPE_CHARACTER);
	if(num_ents != 0)
		return;

	int bid = bot_can_spawn();
	if(bid == -1) return;
	
	game.players[bid] = new(bid) PLAYER(bid);
	game.players[bid]->make_bot(spawn);
	spawn->spawned = true;
	
	if(spawn->level == 4)
	{
		boss.client_id = bid;
		boss.shield_active = false;
		boss.shield_timer = (float)server_tick() - 1;
	}
}


void GAMECONTROLLER_EXP::remove_bot(int id, bool killed)
{
	game.players[id]->bot_spawn->spawned = false;
	if(killed)
		game.players[id]->bot_spawn->respawn_timer = (float)server_tick();
	
	if(killed)
	{
		PICKUP * pickup = new PICKUP(0, 0);
		pickup->pos = game.players[id]->character->pos;
		pickup->create_random_from_bot(game.players[id]->bot_level);
	}
	
	delete game.players[id]->character;
	game.players[id]->character = NULL;
	delete game.players[id];
	game.players[id] = NULL;
}
