#include <engine/e_server_interface.h> //server_tick, server_tickspeed, time_get
#include <engine/e_config.h> //sv_max_clients

#include <game/collision.hpp>
#include <game/server/gamecontext.hpp>
#include <game/server/entities/laser.hpp>
#include <game/server/entities/door.hpp>
#include <game/server/entities/projectile.hpp>
#include <game/server/entities/pickup.hpp>
#include "exp.hpp"



void GAMECONTROLLER_EXP::tick_weapons()
{
	//TELEPORTS
	for(int id = 0; id < config.sv_max_clients; id++)
	{
		if(game.players[id] && game.players[id]->get_character())
		{
			vec2 dest = col_teleport(game.players[id]->get_character()->pos.x, game.players[id]->get_character()->pos.y);
			if(dest.x >= 0 && dest.y >= 0)
			{
				game.players[id]->get_character()->core.pos = dest;
				game.players[id]->get_character()->core.vel = vec2(0, 0);
			}
		}
	}
	
	//HEALING AND GAS ZONES
	for(int id = 0; id < config.sv_max_clients; id++)
	{
		if(game.players[id] && game.players[id]->get_character())
		{
			if(col_get(game.players[id]->get_character()->pos.x, game.players[id]->get_character()->pos.y) & COLFLAG_HEALING)
			{
				if((float)server_tick() > game.players[id]->game_exp.regen_timer)
				{
					if(game.players[id]->get_character()->health < 10)
					{
						game.players[id]->get_character()->health++;
						game.players[id]->game_exp.regen_timer = (float)server_tick() + tuning.regen_timer*server_tickspeed();
					}
					else if(game.players[id]->get_character()->armor < game.players[id]->game_exp.armor_max)
					{
						game.players[id]->get_character()->armor++;
						game.players[id]->game_exp.regen_timer = (float)server_tick() + tuning.regen_timer*server_tickspeed();
					}
				}
			}
			
			if(col_get(game.players[id]->get_character()->pos.x, game.players[id]->get_character()->pos.y) & COLFLAG_POISON)
			{
				if((float)server_tick() > game.players[id]->game_exp.poison_timer)
				{
					game.players[id]->get_character()->take_damage(vec2(0, 0), 1, id, WEAPON_WORLD);
					game.players[id]->game_exp.poison_timer = (float)server_tick() + tuning.poison_timer*server_tickspeed();
				}
			}
		}
	}
	
	//TURRETS
	for(int t = 0; t < cur_turret; t++)
	{
		if(!turrets[t].used)
			continue;
		
		if(turrets[t].dead)
		{
			if((float)server_tick() > turrets[t].timer_respawn + tuning.respawn_timer*server_tickspeed())
				build_turret(t);
			continue;
		}
		
		if(turrets[t].frozen)
		{
			if((float)server_tick() > turrets[t].frozen_timer)
				turrets[t].frozen = false;
		}
		
		//LASER TURRETS
		if(turrets[t].type == TURRET_TYPE_LASER)
		{
			new LASER(turrets[t].pos, vec2(0, 0), 0.0f, -1, true, false);
			
			if(server_tick() < turrets[t].timer || turrets[t].frozen)
				continue;
			
			CHARACTER * close_characters[MAX_CLIENTS];
			int num = game.world.find_entities(turrets[t].pos, tuning.turret_laser_radius, (ENTITY**)close_characters, MAX_CLIENTS, NETOBJTYPE_CHARACTER);
			for(int i = 0; i < num; i++)
			{
				if(!close_characters[i] || close_characters[i]->player->is_bot() || col_intersect_line(turrets[t].pos, close_characters[i]->pos, NULL, NULL, false)) 
					continue;
				
				vec2 direction = normalize(close_characters[i]->pos-turrets[t].pos);
				new LASER(turrets[t].pos, direction, tuning.turret_laser_radius, -1, true, false);
				turrets[t].timer = (float)server_tick() + tuning.turret_reload*server_tickspeed();
				game.create_sound(turrets[t].pos, SOUND_RIFLE_FIRE);
				
				break;
			}
		}
		
		//GUN TURRETS
		else if(turrets[t].type == TURRET_TYPE_GUN)
		{
			if(server_tick() < turrets[t].timer || turrets[t].frozen)
				continue;
			
			CHARACTER * close_characters[MAX_CLIENTS];
			int num = game.world.find_entities(turrets[t].pos, 1000, (ENTITY**)close_characters, MAX_CLIENTS, NETOBJTYPE_CHARACTER);
			for(int i = 0; i < num; i++)
			{
				if(!close_characters[i] || close_characters[i]->player->is_bot() || col_intersect_line(turrets[t].pos, close_characters[i]->pos, NULL, NULL, false)) 
					continue;
				
				vec2 direction = normalize(close_characters[i]->pos-turrets[t].pos);
				new PROJECTILE(WEAPON_GUN, -1, turrets[t].pos, direction, (int)(server_tickspeed()*tuning.gun_lifetime), 1, 0, 0, -1, WEAPON_GUN);
				turrets[t].timer = (float)server_tick() + tuning.turret_reload*server_tickspeed();
				game.create_sound(turrets[t].pos, SOUND_GUN_FIRE);
				
				break;
			}
		}
	}
	
	
	//MINES
	for(int m = 0; m < cur_mine; m++)
	{
		if(!mines[m].used)
			continue;
		
		if(mines[m].dead)
		{
			if((float)server_tick() > mines[m].timer_respawn + tuning.respawn_timer*server_tickspeed())
				build_mine(m);
		}
		else
		{
			//create tic-tic
			CHARACTER * closest = game.world.closest_character_nobot(mines[m].pos, 400);
			if(closest)
			{
				int mod = (int)( distance(closest->pos, mines[m].pos)/8 );
				if(mod == 0 || server_tick()%mod == 0)
					game.create_sound(mines[m].pos, SOUND_HOOK_ATTACH_GROUND);
			}
			
			//emote close players
			CHARACTER * close_chars[MAX_CLIENTS];
			int n = game.world.find_entities(mines[m].pos, 300.0f, (ENTITY**)close_chars, MAX_CLIENTS, NETOBJTYPE_CHARACTER);
			for(int i = 0; i < n; i++)
			{
				if(!close_chars[i] || close_chars[i]->player->is_bot() || col_intersect_line(mines[m].pos, close_chars[i]->pos, NULL, NULL, false)) 
					continue;
				close_chars[i]->emote_type = EMOTE_SURPRISE;
				close_chars[i]->emote_stop = server_tick() + 1*server_tickspeed();
				if(time_get() > close_chars[i]->player->last_emote+time_freq()*3)
				{
					game.send_emoticon(close_chars[i]->player->client_id, EMOTICON_2); //!
					close_chars[i]->player->last_emote = time_get();
				}
			}
			
			CHARACTER * close_characters[MAX_CLIENTS];
			int num = game.world.find_entities(mines[m].pos, 16.0f, (ENTITY**)close_characters, MAX_CLIENTS, NETOBJTYPE_CHARACTER);
			for(int i = 0; i < num; i++)
			{
				if(!close_characters[i] || close_characters[i]->player->is_bot() || col_intersect_line(mines[m].pos, close_characters[i]->pos, NULL, NULL, false)) 
					continue;
				destroy_mine(m);
				break;
			}
		}
	}
	
	
	//TRAPS
	for(int t = 0; t < cur_trap; t++)
	{
		if(!traps[t].used)
			continue;
		if((float)server_tick() < traps[t].timer)
			continue;
		
		CHARACTER * close_characters[MAX_CLIENTS];
		int num = game.world.find_entities(traps[t].pos, 600.0f, (ENTITY**)close_characters, MAX_CLIENTS, NETOBJTYPE_CHARACTER);
		for(int i = 0; i < num; i++)
		{
			if(!close_characters[i] || close_characters[i]->player->is_bot() || col_intersect_line(traps[t].pos, close_characters[i]->pos, NULL, NULL, false)) 
				continue;
			if(close_characters[i]->pos.y > traps[t].pos.y && close_characters[i]->pos.x > traps[t].pos.x-64 && close_characters[i]->pos.x < traps[t].pos.x+64)
			{
				new PROJECTILE(WEAPON_GRENADE, -1, traps[t].pos, vec2(0, 1), (int)(server_tickspeed()*tuning.grenade_lifetime), 1, PROJECTILE::PROJECTILE_FLAGS_EXPLODE, 0, SOUND_GRENADE_EXPLODE, WEAPON_GRENADE);
				traps[t].timer = (float)server_tick() + data->weapons.id[WEAPON_GRENADE].firedelay*server_tickspeed()/1000.0f;
				game.create_sound(traps[t].pos, SOUND_GRENADE_FIRE);
				break;
			}
		}
	}
	
	
	//DOORS
	for(int d = 0; d < cur_door; d++)
	{
		if(!doors[d].used)
			continue;
		
		if(doors[d].create_laser)
		{
			doors[d].laser = new LASER_DOOR(doors[d].pos, doors[d].type, &doors[d]);
			doors[d].create_laser = false;
		}
	}
}



void GAMECONTROLLER_EXP::build_turret(int t)
{
	turrets[t].dead = false;
	turrets[t].life = tuning.turret_life;
	turrets[t].timer = 0.0f;
	turrets[t].timer_respawn = 0.0f;
	turrets[t].frozen = false;
	game.create_playerspawn(turrets[t].pos);
}

void GAMECONTROLLER_EXP::hit_turret(int t, int owner, int dmg)
{
	//create damage indicator :
	turrets[t].damage_taken += dmg;
	if(server_tick() < turrets[t].damage_taken_tick+25)
		game.create_damageind(turrets[t].pos, turrets[t].damage_taken*0.25f, dmg);
	else
	{
		turrets[t].damage_taken = 0;
		game.create_damageind(turrets[t].pos, 0, dmg);
	}
	turrets[t].damage_taken_tick = server_tick();
	
	game.create_sound(game.players[owner]->view_pos, SOUND_HIT, cmask_one(owner));
	
	turrets[t].life -= dmg;
	if(turrets[t].life <= 0)
		destroy_turret(t, owner);
	char buff[255];
	str_format(buff, sizeof(buff), "Turret [Health: %i]", turrets[t].life);
	game.send_broadcast(buff, owner);
}

void GAMECONTROLLER_EXP::destroy_turret(int t, int killer)
{
	turrets[t].dead = true;
	turrets[t].life = 0;
	turrets[t].timer = 0.0f;
	turrets[t].timer_respawn = (float)server_tick();
	
	game.players[killer]->score++;
	
	if(server_tick()%100 < 75)
	{
		PICKUP * pickup = new PICKUP(0, 0);
		pickup->pos = turrets[t].pos;
		pickup->create_random_from_turret(turrets[t].type);
	}
	
	game.create_death(turrets[t].pos, config.sv_max_clients);
}

void GAMECONTROLLER_EXP::freeze_turret(int t)
{
	turrets[t].frozen = true;
	turrets[t].frozen_timer = (float)server_tick() + 5.0f*server_tickspeed();
}


void GAMECONTROLLER_EXP::build_mine(int m)
{
	mines[m].dead = false;
	mines[m].timer_respawn = 0.0f;
}

void GAMECONTROLLER_EXP::destroy_mine(int m)
{
	mines[m].dead = true;
	mines[m].timer_respawn = (float)server_tick();
	
	game.create_explosion(mines[m].pos, -1, WEAPON_WORLD, false);
	game.create_sound(mines[m].pos, SOUND_GRENADE_EXPLODE);
}


void GAMECONTROLLER_EXP::build_trap(int t)
{
	traps[t].timer = 0.0f;
}


void GAMECONTROLLER_EXP::build_door(int d)
{
	doors[d].create_laser = false;
	doors[d].laser = new LASER_DOOR(doors[d].pos, doors[d].type, &doors[d]);
	col_set_door(doors[d].laser->from.x, doors[d].laser->from.y, doors[d].laser->pos.x, doors[d].laser->pos.y);
}
